/*********************/
/* errmsg.c          */
/* for Par 1.52      */
/* Copyright 2001 by */
/* Adam M. Costello  */
/*********************/

/* This is ANSI C code (C89). */


#include "errmsg.h"  /* Makes sure we're consistent with the declarations. */


const char * const outofmem =
  "Out of memory.\n";

const char * const impossibility =
  "Impossibility #%d has occurred.  Please report it.\n";
